/*
 * tmos - thread.c
 * 	version 0.0.0
 *
 * 	TMOS multitasking demo
 * 
 * Copyright (C) 2005..2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <asm/system.h>
#include <tmos/sched.h>
#include <tmos/port.h>
#include <tmos/mutex.h>

/*
 * context for task A
 */
static struct context __kernel task_a;

/* console port */
static port_t port;

/* mutex to protect console from concurent access */
static DECLARE_MUTEX(put_lock);

/**
 * __putc - print character to console
 * @c: character to print
 *
 * This function must be called from kernel mode with console lock held.
 */
static void __putc(char c) __reentrant
{
	static char __user t;

	t = c;
	sys_port_write(port, &t, 1);
}

/**
 * putc - print character to console
 * @c: character to print
 *
 * This function must be called from user mode.
 */
static void putc(char c) __reentrant
{
	kernel_enter();
	mutex_lock(put_lock);
	__putc(c);
	mutex_unlock(put_lock);
	kernel_exit();
}

/**
 * puts - print string to console
 * @s: string to print
 *
 * This function must be called from user mode.
 */
static void puts(const char *s) __reentrant
{
	while (*s)
		putc(*s++);
}

/**
 * task_a_f - task A code
 */
void task_a_f(void) __reentrant
{
	tsc_t ht;

	/*
	 * thread_create must be called in kernel mode with disabled
	 * preemption
	 */
	kernel_enter();
	preempt_disable();
	if (!thread_create(&task_a, TASK_STACK())) {
		/* parent */
		preempt_enable();

		/* new threads are STOPPED and should be woken up */
		wake_up_context(&task_a);

		kernel_exit();
		return;
	}
	/*
	 * child is called in kernel mode, but preemption is ENABLED!
	 * It's because preemption in TMOS is global, not per-task
	 * as in Linux.
	 */
	kernel_exit();
		
	/* child code */
	for (;;) {
		ht = tsc_read();
		kernel_enter();
		schedule();
		kernel_exit();
		ht = tsc_read() - ht;

		kernel_enter();
		printk("sched: %d\n", ht);
		kernel_exit();
	}
}

void app_main(void) __reentrant
{
	kernel_enter();
	printk("openning console\n");
	port = sys_port_open(CONFIG_CONSOLE_DEVICE, O_RDWR);
	kernel_exit();
	if (port == -1)
		BUG();
	
	puts("testing puts.\n");

	kernel_enter();
	printk("creating task A\n");
	kernel_exit();
	task_a_f();

	for (;;) {
		kernel_enter();
		schedule();
		kernel_exit();
	}
}
