/*
 * tmos - i51.c
 *
 * SLIP for i51 serial
 * 
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/port.h>
#include <tmos/crc16.h>
#include <asm/8052.h>
#include <asm/system.h>
#include <asm/task.h>
#include <tmos/errno.h>
#include <tmos/net.h>
#include <asm/irqs.h>
#include <limits.h>

#ifndef dir_tx
#define dir_tx()	do { } while (0)
#endif
#ifndef dir_rx
#define dir_rx()	do { } while (0)
#endif

#define serial_interrupt_disable() do { ES = 0; } while (0)
#define serial_interrupt_enable() do { ES = 1; } while (0)

/* SLIP protocol characters. */
#define END             0300		/* indicates end of frame	*/
#define ESC             0333		/* indicates byte stuffing	*/
#define ESC_END         0334		/* ESC ESC_END means END 'data'	*/
#define ESC_ESC         0335		/* ESC ESC_ESC means ESC 'data'	*/

#define SLIP_MTU		CONFIG_SLIP_MTU

#define TXP_END 0xff

static uint8_t __kernel msg_buffer[SLIP_MTU];
static uint8_t __kernel msg_size = 0;
static uint8_t __kernel msg_txp = TXP_END;	/* transmit pointer */
static __bit rx_lock = 0;
static __bit esc = 0;

#ifdef CONFIG_SLIP_ADDR_FILTER

#ifdef CONFIG_SLIP_ADDR_FILTER_RUNTIME
static uint8_t slip_addr = CONFIG_SLIP_ADDR_FILTER_ADDR;
#define slip_address()	slip_addr
#else
#define slip_address()	CONFIG_SLIP_ADDR_FILTER_ADDR
#endif

#ifdef CONFIG_SLIP_ADDR_FILTER_BROADCAST
#define msg_filter() \
	((msg_buffer[0] == slip_address()) || !msg_buffer[0])
#else
#define msg_filter() (msg_buffer[0] == slip_address())
#endif
#else
#define msg_filter() (1)
#endif

static inline void tx_send_slip_raw(unsigned char x)
{
	DPL = x;
	__asm
		mov	a, _DPL
#if defined (CONFIG_SERIAL_PARITY_EVEN)
		mov	c, _P
		mov	_TB8, c
#elif defined (CONFIG_SERIAL_PARITY_ODD)
		mov	c,_P
		cpl	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_SPACE)
		clr	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_MARK)
		setb	c	
		mov	_TB8,c
		
#endif		
		mov	SBUF, a
	__endasm;
}

static inline void tx_send(unsigned char x)
{
	if (esc) {
		esc = 0;
		if (x == ESC)
			x = ESC_ESC;
		else
			x = ESC_END;
	} else {
		if (x == END)
			x = ESC;
		if (x == ESC)
			esc = 1;
	}

	if (!esc)
		msg_txp++;
	tx_send_slip_raw(x);
}

void ser_irq(void) __interrupt SERIAL_INTERRUPT __using I51_SERIAL_BANK
{
	unsigned char tmp;
	
	if (RI) {
		RI = 0;
		tmp = SBUF;

		if (!rx_lock) {
			switch (tmp) {
			case END:
				rx_lock = 1;
				/* fall trough */
			case ESC:
				esc = 1;
				break;
			case ESC_ESC:
				if (esc) {
					esc = 0;
					tmp = ESC;
				}
				break;
			case ESC_END:
				if (esc) {
					esc = 0;
					tmp = END;
				}
				break;
			}

			if (!esc) {
				if (msg_size < SLIP_MTU) {
					msg_buffer[msg_size++] = tmp;
				} else {
					msg_size = SLIP_MTU + 1;
				}
			} else if (rx_lock) {
				if ((msg_size >= 2) && (msg_size <= SLIP_MTU)
						&& msg_filter()) {
					msg_txp = TXP_END;
				} else {
					rx_lock = 0;
					msg_size = 0;
				}

				esc = 0;
			}
		}
	}

	if (TI) {
		TI = 0;
		if (msg_txp < msg_size) {
			tx_send(msg_buffer[msg_txp]);
		} else if (msg_txp == msg_size) {
			tx_send_slip_raw(END);
			msg_txp++;
		} else {
			dir_rx();
			msg_size = 0;
			rx_lock = 0;
		}
	}
}

char slip_i51_ioctl(unsigned char cmd, void __kernel *arg) __reentrant
{
	char ret = -EINVAL;

	switch (cmd) {
#ifdef CONFIG_SLIP_ADDR_FILTER_RUNTIME
	case NET_IOCTL_SET_ADDR_BYTE:
		slip_addr = *((uint8_t __kernel *) arg);
		ret = 0;
		break;
#endif
	case NET_IOCTL_SET_SPEED:
		ret = 0;
		switch (*((uint8_t __kernel *) arg)) {
		case NET_SPEED_9600:
			ser_set_spd(B9600);
		default:
			ret = -EINVAL;
		}
		break;
	}
	return ret;
}

void slip_i51_poll(void) __reentrant
{
	if (rx_lock && (msg_txp == TXP_END)) {
		net_size_t ret = 0;
		uint16_t __kernel *crc;
		struct net_msg msg;

		kernel_enter();

		/* setup message */
		msg_init(&msg);		
		msg_set_interface(&msg, NET_IFACE_SERIAL0);
		msg_set_buffer(&msg, msg_buffer);
		msg_set_head(&msg, 0);

#if CONFIG_SLIP_CRC16
		msg_set_tail(&msg, msg_size - 2);
		msg_set_size(&msg, SLIP_MTU - 2);
		crc = (uint16_t __kernel *)(msg_buffer + msg_size - 2);
		if (le16_to_cpu(*crc) == crc16_kernel(0xffff, msg_buffer,
					msg_size - 2)) {
			ret = net_xmit(NET_PROTO_MODBUS, &msg);
			if (ret) {
				msg_size = ret + 2;
				crc = (uint16_t __kernel *)(msg_buffer + ret);
				*crc = cpu_to_le16(crc16_kernel(0xffff,
							msg_buffer, ret));
			}
		}
#else
		msg_set_tail(&msg, msg_size);
		msg_set_size(&msg, SLIP_MTU);
		ret = net_xmit(NET_PROTO_MODBUS, &msg);
		if (ret)
			msg_size = ret;
#endif

		serial_interrupt_disable();
		if (ret) {
			dir_tx();
			msg_txp = 0;

			tx_send_slip_raw(END);
		} else {
			msg_size = 0;
			rx_lock = 0;
		}
		serial_interrupt_enable();
		kernel_exit();
	}
}

static void slip_i51_init(void) __reentrant
{
	dir_rx();
}

device_initcall(slip_i51_init)
