/*
 * tmos - i51.c
 * 
 * Copyright (C) 2005..2009  Krzysztof Mazur <krzysiek@podlesie.net>
 * Copyright (C) 2009  Marek Piwonski <marek@microster.pl>	
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/port.h>
#include <asm/8052.h>
#include <asm/system.h>
#include <tmos/sched.h>
#include <tmos/errno.h>
#include <limits.h>

#ifdef CONFIG_SERIAL_TIMER3 
#include <asm/aduc841.h>
#endif

#ifdef CONFIG_SERIAL_I51_STREAM

#define RX_RING_SIZE	(1 << CONFIG_SERIAL_I51_RX_BUFFER_SHIFT)
#define RX_RING_MASK	(RX_RING_SIZE - 1)
#define TX_RING_SIZE	(1 << CONFIG_SERIAL_I51_TX_BUFFER_SHIFT)
#define TX_RING_MASK	(TX_RING_SIZE - 1)

static uint8_t __kernel rx_ring[RX_RING_SIZE];
static uint8_t __kernel tx_ring[TX_RING_SIZE];
uint8_t __kernel rx_head = 0;
uint8_t __kernel rx_tail = 0;
uint8_t __kernel tx_head = 0;
uint8_t __kernel tx_tail = 0;

/**
 * ser_rx_wait - wait for character to receive
 * 
 * This function should be called in kernel mode with interrupts enabled.
 */
void ser_rx_wait(void) __kernel_preempt
{
	while (rx_head == rx_tail)
		schedule();
}

/**
 * __ser_rxc - receive single character
 *
 * There MUST be a character in receive buffer. This function is not
 * reentrant - should be called in kernel mode with preemption disabled.
 */
static char __ser_rxc(void) __kernel_preempt
{
	char ret;

	ret = rx_ring[rx_head];
	rx_head = (rx_head + 1) & RX_RING_MASK;
	return ret;
}

/**
 * ser_rxc - receive single character
 */
char ser_rxc(void) __user_preempt
{
	char ret;

	ser_rx_wait();
	ret = __ser_rxc();
	return ret;
}

static inline void tx_send(void)
{
#ifdef CONFIG_KERNEL_DATA
	__asm
		mov	a, #_tx_ring
		add	a, _tx_head
		mov	r0, a
		mov	a, @r0
#if defined (CONFIG_SERIAL_PARITY_EVEN)
		mov	c, _P
		mov	_TB8, c
#elif defined (CONFIG_SERIAL_PARITY_ODD)
		mov	c,_P
		cpl	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_SPACE)
		clr	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_MARK)
		setb	c	
		mov	_TB8,c
		
#endif		
		mov	SBUF, a
	__endasm;
#elif CONFIG_KERNEL_XDATA
	__asm
		mov	dptr, #_tx_head
		movx	a, @dptr
		add	a, #_tx_ring
		mov	dpl, a
		clr	a
		addc	a, #(_tx_ring >> 8)
		mov	dph, a
		movx	a, @dptr
#if defined (CONFIG_SERIAL_PARITY_EVEN)
		mov	c, _P
		mov	_TB8, c
#elif defined (CONFIG_SERIAL_PARITY_ODD)
		mov	c,_P
		cpl	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_SPACE)
		clr	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_MARK)
		setb	c	
		mov	_TB8,c
		
#endif		
		mov	SBUF, a
	__endasm;
#else
#warning "unsupported __kernel memory"
	SBUF = tx_ring[tx_head];
#endif
}

static __bit transmiter_ready = 1;

/*
 * ser_txc - transmit character
 * @c: character
 *
 * This function should be called from kernel context with interrupts
 * enabled.
 */
void ser_txc(char c) __kernel_preempt
{
	uint8_t tmp;

	for (;;) {
		preempt_disable();
		tmp = (tx_tail + 1) & TX_RING_MASK;
		if (tmp != tx_head)
			break;
		preempt_enable();
		schedule();
	}
	tx_ring[tx_tail] = c;
	smp_wmb();
	tx_tail = tmp;

	if (transmiter_ready) {
		transmiter_ready = 0;
		smp_wmb();
		tx_send();
	}
	preempt_enable();
}


#ifdef CONFIG_SERIAL_DATA
char ser_rxd(uint8_t __data *buf, char len) REENTRANT
{
	char __data *t;
	char tmp;

	t = (char __data *) buf;
	ser_rx_wait();
	for (tmp = 0; tmp < len && ser_rx_poll(); tmp++, t++)
		*t = __ser_rxc();
	return tmp;
}

#if CONFIG_SERIAL_DEPRECATED
char ser_rxd_t(uint8_t __data *buf, char len, jiffies_t jiffies) REENTRANT
{
	char __data *t;
	char tmp;
	jiffies_t t0;

	t = (char __data *) buf;
	for (tmp = 0; tmp < len; tmp++, t++) {
		t0 = get_jiffies();
		while (!ser_rx_poll()) {
			if (get_jiffies() - t0 > jiffies)
				return tmp;
			schedule();
		}
		*t = __ser_rxc();
	}
	return tmp;
}
#endif

void ser_txd(const uint8_t __data *buf, char len) REENTRANT
{
	while (len--)
		ser_txc(*buf++);
}
#endif

#ifdef CONFIG_SERIAL_XDATA
char ser_rxxd(uint8_t __xdata *buf, char len) REENTRANT
{
	char __xdata *t;
	char tmp;

	t = (char __xdata *) buf;
	ser_rx_wait();
	for (tmp = 0; tmp < len && ser_rx_poll(); tmp++, t++)
		*t = __ser_rxc();
	return tmp;
}

#if CONFIG_SERIAL_DEPRECATED
char ser_rxxd_t(uint8_t __xdata *buf, char len, jiffies_t jiffies) REENTRANT
{
	char __xdata *t;
	char tmp;
	jiffies_t t0;

	t = (char __xdata *) buf;
	for (tmp = 0; tmp < len; tmp++, t++) {
		t0 = get_jiffies();
		while (!ser_rx_poll()) {
			if (get_jiffies() - t0 > jiffies)
				return tmp;
			schedule();
		}
		*t = __ser_rxc();
	}
	return tmp;
}
#endif

void ser_txxd(const uint8_t __xdata *buf, char len) REENTRANT
{
	while (len--)
		ser_txc(*buf++);
}
#endif

#ifdef CONFIG_SERIAL_GENERIC
char ser_rx(uint8_t *buf, char len) REENTRANT
{
	char *t;
	char tmp;

	t = (char *) buf;
	ser_rx_wait();
	for (tmp = 0; tmp < len && ser_rx_poll(); tmp++, t++)
		*t = __ser_rxc();
	return tmp;
}

#if CONFIG_SERIAL_DEPRECATED
char ser_rx_t(uint8_t *buf, char len, jiffies_t jiffies) REENTRANT
{
	char *t;
	char tmp;
	jiffies_t t0;

	t = (char *) buf;
	for (tmp = 0; tmp < len; tmp++, t++) {
		t0 = get_jiffies();
		while (!ser_rx_poll()) {
			if (get_jiffies() - t0 > jiffies)
				return tmp;
			schedule();
		}
		*t = __ser_rxc();
	}
	return tmp;
}
#endif
#endif

void ser_tx(const uint8_t *buf, char len) REENTRANT
{
	while (len--)
		ser_txc(*buf++);
}

void ser_txs(const char *str) __user_preempt
{ 
	while (*str != '\0')
		ser_txc(*str++);
}

void ser_irq(void) __interrupt 4 __using I51_SERIAL_BANK
{
	unsigned char tmp;

	if (RI) {
		RI = 0;
		tmp = rx_tail;
	       	tmp++;
		tmp &= RX_RING_MASK;
		if (tmp != rx_head) {
			rx_ring[rx_tail] = SBUF;
			rx_tail = tmp;
		} else {
			tmp = SBUF;
		}
	}

	if (TI) {
		TI = 0;
		tx_head++;
		tx_head &= TX_RING_MASK;
		if (tx_head != tx_tail) {
			tx_send();
		} else {
			transmiter_ready = 1;
		}
	}

	set_need_wakeup();
}

void ser_flush(void) __user_preempt
{
	while (!transmiter_ready)
		schedule();
}
#else

void ser_txc(char c) __kernel_preempt
{
	TI = 0;
	SBUF = c;
	while (!TI);
}

#ifdef CONFIG_SERIAL_SEND_RAW
void tx_send_raw(unsigned char x) __reentrant
{
	DPL = x;
	__asm
		mov	a, _DPL
#if defined (CONFIG_SERIAL_PARITY_EVEN)
		mov	c, _P
		mov	_TB8, c
#elif defined (CONFIG_SERIAL_PARITY_ODD)
		mov	c,_P
		cpl	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_SPACE)
		clr	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_MARK)
		setb	c	
		mov	_TB8,c
		
#endif		
		mov	SBUF, a
	__endasm;
}

void __tx_send_raw(unsigned char x) __using I51_SERIAL_BANK __reentrant
{
	DPL = x;
	__asm
		mov	a, _DPL
#if defined (CONFIG_SERIAL_PARITY_EVEN)
		mov	c, _P
		mov	_TB8, c
#elif defined (CONFIG_SERIAL_PARITY_ODD)
		mov	c,_P
		cpl	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_SPACE)
		clr	c	
		mov	_TB8,c
#elif defined (CONFIG_SERIAL_PARITY_MARK)
		setb	c	
		mov	_TB8,c
		
#endif		
		mov	SBUF, a
	__endasm;
}
#endif

#endif

void ser_set_spd(uint16_t spd) REENTRANT
{
#if defined(CONFIG_SERIAL_TIMER3)
	if (spd == B9600) {
		T3CON = 0x86;
		T3FD = 0x08;
	} else if (spd == B19200) {
		T3CON = 0x85;
		T3FD = 0x08;
	} else if (spd == B38400) {
		T3CON = 0x84;
		T3FD = 0x08;
	} else if (spd == B57600) {
		T3CON = 0x83;
		T3FD = 0x20;
	} else if (spd == B115200) {
		T3CON = 0x82;
		T3FD = 0x20;
	} else if (spd == B125000) {
		T3CON = 0x82;
		T3FD = 24;
	} else if (spd == B150000) {
		T3CON = 0x82;
		T3FD = 10;
	} else if (spd == B230400) {
		T3CON = 0x81;
		T3FD = 0x20;
	} else {
		T3CON = 0x86;
		T3FD = 0x08;
	}
#elif defined(CONFIG_USE_BRG)
	BRL = spd;
#elif defined(CONFIG_SERIAL_TIMER2)
	TL2 = RCAP2L = spd & 0xff;
	TH2 = RCAP2H = spd >> 8;
#else
	TH1 = spd;
#endif
}

static void i51_serial_init(void)
{
	PCON |= SMOD;

#if defined (CONFIG_SERIAL_PARITY_NONE)
	SCON = 0x50;
#else
	SCON = 0xd0;
#endif

#if defined(CONFIG_SERIAL_TIMER3)
#elif defined(CONFIG_USE_BRG)
	BDRCON |= BDRCON_BRR | BDRCON_TBCK | BDRCON_RBCK | BDRCON_SPD
		| BDRCON_SRC;
#elif defined(CONFIG_SERIAL_TIMER2)
	T2CON = 0x30;
#else
	TMOD = (TMOD & 0xf0) | 0x20;
#endif
	ser_set_spd(BDEFAULT);
	
#if defined(CONFIG_SERIAL_TIMER3)
#elif defined(CONFIG_USE_BRG)
#elif defined(CONFIG_SERIAL_TIMER2)
	TR2 = 1;
#else
	TR1 = 1;
#endif
#if CONFIG_I51_SERIAL_HIGHPRIO
	PS = 1;
#endif

	RI = 0;	
	TI = 0;
	ES = 1;
}

#if CONFIG_SERIAL_I51_STREAM && CONFIG_PORT
msg_ssize_t serial_i51_xmit(port_t port, char __user *buf, msg_size_t len,
		msg_size_t max_resp_len, msg_flags_t flags, off_t off)
	__reentrant
{
	msg_ssize_t ret;

#if !CONFIG_PORT_MSG_SIZE_CHAR
	if (len > CHAR_MAX)
		len = CHAR_MAX;
#endif
	ret = len;
	/* FIXME: don't wait in write() if MST_DONTWAIT is set */
	ser_tx_user(buf, len);
	if (max_resp_len) {
		if ((flags & MSG_DONTWAIT) && !ser_rx_poll())
			return -EAGAIN;
#if !CONFIG_PORT_MSG_SIZE_CHAR
		if (max_resp_len > CHAR_MAX)
			max_resp_len = CHAR_MAX;
#endif
		ret = ser_rx_user(buf, max_resp_len);
	}
	return ret;
}
#endif

arch_initcall(i51_serial_init)
