/*
 * tmos - mb-slave.c
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * TinyRT Modbus slave support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <tmos/kernel.h>
#include <tmos/trt.h>
#include <tmos/mb-slave.h>
#include <tmos/modbus.h>
#include <tmos/port.h>

#if CONFIG_TRT_MB_SLAVE_READI_INPUTBASE >= TRT_REGS
#error "invalid read_input inputbase"
#endif

#if CONFIG_TRT_MB_SLAVE_READ_INPUTBASE >= TRT_REGS
#error "invalid read inputbase"
#endif

#if CONFIG_TRT_MB_SLAVE_WRITE_INPUTBASE >= TRT_REGS
#error "invalid write inputbase"
#endif

#define MAX_READI_REGS (TRT_REGS - CONFIG_TRT_MB_SLAVE_READI_INPUTBASE)
#define MAX_READ_REGS (TRT_REGS - CONFIG_TRT_MB_SLAVE_READ_INPUTBASE)
#define MAX_WRITE_REGS (TRT_REGS - CONFIG_TRT_MB_SLAVE_WRITE_INPUTBASE)

#if (CONFIG_TRT_MB_SLAVE_READI_COUNT < 0) || (CONFIG_TRT_MB_SLAVE_READI_COUNT > MAX_READI_REGS)
#define READI_LIMIT MAX_READI_REGS
#else
#define READI_LIMIT CONFIG_TRT_MB_SLAVE_READI_COUNT
#endif

#if (CONFIG_TRT_MB_SLAVE_READ_COUNT < 0) || (CONFIG_TRT_MB_SLAVE_READ_COUNT > MAX_READ_REGS)
#define READ_LIMIT MAX_READ_REGS
#else
#define READ_LIMIT CONFIG_TRT_MB_SLAVE_READ_COUNT
#endif

#if (CONFIG_TRT_MB_SLAVE_WRITE_COUNT < 0) || (CONFIG_TRT_MB_SLAVE_WRITE_COUNT > MAX_WRITE_REGS)
#define WRITE_LIMIT MAX_WRITE_REGS
#else
#define WRITE_LIMIT CONFIG_TRT_MB_SLAVE_WRITE_COUNT
#endif

#ifdef CONFIG_TRT_MB_SLAVE_MAP
static port_t map_port = -1;

#define MAPPING_START	CONFIG_TRT_MB_SLAVE_MAP_BASE
#define MAPPING_SIZE	CONFIG_TRT_MB_SLAVE_MAP_SIZE
#define MAPPING_END	(MAPPING_START + MAPPING_SIZE)
#endif

mb_ex_t mb_slave_hr_read(uint16_t __kernel *v, mb_addr_t addr, mb_size_t len)
	__reentrant
{
	unsigned char count = READ_LIMIT;

#ifdef CONFIG_TRT_MB_SLAVE_MAP
	if ((addr >= MAPPING_START) && (addr < MAPPING_END)) {
		static uint8_t __user buffer[32];

		addr -= MAPPING_START;
		if ((addr >= MAPPING_SIZE) || (addr + len > MAPPING_SIZE)) {
			return MBEX_FAULT;
		}

		if (map_port == -1)
			return MBEX_FAULT;

		if (2 * len > 16)
			return MBEX_FAULT;

		sys_port_pread(map_port, buffer, 2 * len, 2 * addr);
		copy_from_user(v, buffer, 2 * len);
		return 0;
	}
#endif

	if ((addr >= count) || (addr + len > count))
		return MBEX_FAULT;
	addr += CONFIG_TRT_MB_SLAVE_READ_INPUTBASE;
	while (len > 0) {
		*v = cpu_to_be16(trt_read_reg(addr));
		v++;
		addr++;
		len--;
	}
	return 0;
}

mb_ex_t mb_slave_hr_write(const uint16_t __kernel *v, mb_addr_t addr,
		mb_size_t len) __reentrant
{
	unsigned char count = READ_LIMIT;

#ifdef CONFIG_TRT_MB_SLAVE_MAP
	if ((addr >= MAPPING_START) && (addr < MAPPING_END)) {
		static uint8_t __user buffer[32];

		addr -= MAPPING_START;
		if ((addr >= MAPPING_SIZE) || (addr + len > MAPPING_SIZE)) {
			return MBEX_FAULT;
		}

		if (map_port == -1)
			return MBEX_FAULT;

		if (2 * len > 32)
			return MBEX_FAULT;

		copy_to_user(buffer, v, 2 * len);
		if (sys_port_pwrite(map_port, buffer, 2 * len, 2 * addr)
				!= 2 * len)
			return MBEX_FAULT;
		return 0;
	}
#endif

	if ((addr >= count) || (addr + len > count))
		return MBEX_FAULT;
	addr += CONFIG_TRT_MB_SLAVE_WRITE_INPUTBASE;
	while (len > 0) {
		trt_write_reg(be16_to_cpu(*v), addr);
		v++;
		addr++;
		len--;
	}
	return 0;
}

mb_ex_t mb_slave_ir_read(uint16_t __kernel *v, mb_addr_t addr, mb_size_t len)
	__reentrant
{
	unsigned char count = READI_LIMIT;

	if ((addr >= count) || (addr + len > count))
		return MBEX_FAULT;
	addr += CONFIG_TRT_MB_SLAVE_READI_INPUTBASE;
	while (len > 0) {
		*v = cpu_to_be16(trt_read_reg(addr));
		v++;
		addr++;
		len--;
	}
	return 0;
}

#ifdef CONFIG_TRT_MB_SLAVE_MAP
void mb_mapping_init(void)
{
	kernel_enter();
	map_port = sys_port_open(CONFIG_TRT_MB_SLAVE_MAP_PATH, O_RDWR);
	kernel_exit();
}
#endif
